"""Units tests on 2 Datasets to detect potential bugs of the Kern_Rowduction package."""

import os
import pandas as pd
import kern_rowduction.rowduction as krd

################## Tests on the Dataset A ##################

def test_rowduction_0_dataset_a(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_pima_extract.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 0 value of the label of the Dataset A.

    Implemented checks :
    -Check if the rowduction reduces the number of rows from the original input df
    -Check if all "one shot" rowductions and rowduction with a
    higher batch size (nb_rows_memory) than the total number of rows give the results
    -Check if the number of rows decreases when epsilon increases
    -Check if the number of rows of a "separately" rowductioned dataframe is equal or higher
    than the one of a "grouped" rowductioned dataframe
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [0], epsilon, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_12 = krd.rowduct(df, [0], epsilon, nb_rows_memory=50000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_13 = krd.rowduct(df, [0], epsilon, nb_rows_memory=50000, step_activated = False, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_14 = krd.apply_kern_rowduction(df[df["Label"] == 0].reset_index(drop=True), \
        epsilon, nb_rows_memory=50000)
    rowducted_df_14 = rowducted_df_14.append(df[df["Label"] == 1]).reset_index(drop=True)
        # With different values of epsilon
    rowducted_df_15 = krd.rowduct(df, [0], epsilon+0.5, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")           
    rowducted_df_16 = krd.rowduct(df, [0], epsilon+1, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [0], epsilon, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_22 = krd.rowduct(df, [0], epsilon, nb_rows_memory=50000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_23 = krd.rowduct(df, [0], epsilon, nb_rows_memory=50000, step_activated = False, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_24 = krd.apply_kern_rowduction(df[df["Label"] == 0].reset_index(drop=True), \
        epsilon, nb_rows_memory=50000)
    rowducted_df_24 = rowducted_df_24.append(df[df["Label"] == 1]).reset_index(drop=True)
        # With different values of epsilon
    rowducted_df_25 = krd.rowduct(df, [0], epsilon+0.5, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")            
    rowducted_df_26 = krd.rowduct(df, [0], epsilon+1, nb_rows_memory=10000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")  

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)
    assert len(rowducted_df_12) < len(df)
    assert len(rowducted_df_13) < len(df)
    assert len(rowducted_df_14) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_15) <= len(rowducted_df_11)
    assert len(rowducted_df_16) <= len(rowducted_df_15)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_11) == len(rowducted_df_12)
    assert len(rowducted_df_12) == len(rowducted_df_13)
    assert len(rowducted_df_13) == len(rowducted_df_14)
    # In terms of rowducted dataframe
    assert (rowducted_df_11 != rowducted_df_12).any(1).count() == len(rowducted_df_12)
    assert (rowducted_df_12 != rowducted_df_13).any(1).count() == len(rowducted_df_13)
    assert (rowducted_df_13 != rowducted_df_14).any(1).count() == len(rowducted_df_14)

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)
    assert len(rowducted_df_22) < len(df)
    assert len(rowducted_df_23) < len(df)
    assert len(rowducted_df_24) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_25) <= len(rowducted_df_21)
    assert len(rowducted_df_26) <= len(rowducted_df_25)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_21) == len(rowducted_df_22)
    assert len(rowducted_df_22) == len(rowducted_df_23)
    assert len(rowducted_df_23) == len(rowducted_df_24)
    # In terms of rowducted dataframe
    assert (rowducted_df_21 != rowducted_df_22).any(1).count() == len(rowducted_df_22)
    assert (rowducted_df_22 != rowducted_df_23).any(1).count() == len(rowducted_df_23)
    assert (rowducted_df_23 != rowducted_df_24).any(1).count() == len(rowducted_df_24)

    # Check if the dataframes rowductioned "separately" have more or at least the same number
    # of rows than one rowductioned in a "grouped" way
    assert len(rowducted_df_21) >= len(rowducted_df_11)
    assert len(rowducted_df_22) >= len(rowducted_df_12)
    assert len(rowducted_df_23) >= len(rowducted_df_13)   
    assert len(rowducted_df_24) >= len(rowducted_df_14)   

def test_rowduction_1_dataset_a(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_pima_extract.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 1 value of the label of the Dataset A.

    Implemented checks :
    -Check if the rowduction reduces the number of rows from the original input df
    -Check if all "one shot" rowductions and rowduction with a
    higher batch size (nb_rows_memory) than the total number of rows give the results
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [1], epsilon, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_12 = krd.rowduct(df, [1], epsilon, nb_rows_memory=32000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_13 = krd.rowduct(df, [1], epsilon, nb_rows_memory=32000, step_activated = False, \
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_14 = krd.apply_kern_rowduction(df[df["Label"] == 1].reset_index(drop=True),  \
        epsilon, nb_rows_memory=32000)
    rowducted_df_14 = rowducted_df_14.append(df[df["Label"] == 0]).reset_index(drop=True)
        # With different values of epsilon
    rowducted_df_15 = krd.rowduct(df, [1], epsilon+0.5, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")         
    rowducted_df_16 = krd.rowduct(df, [1], epsilon+1, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [1], epsilon, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_22 = krd.rowduct(df, [1], epsilon, nb_rows_memory=32000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_23 = krd.rowduct(df, [1], epsilon, nb_rows_memory=32000, step_activated = False, \
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_24 = krd.apply_kern_rowduction(df[df["Label"] == 1].reset_index(drop=True), \
        epsilon, nb_rows_memory=32000)
    rowducted_df_24 = rowducted_df_24.append(df[df["Label"] == 0]).reset_index(drop=True)
        # With different values of epsilon
    rowducted_df_25 = krd.rowduct(df, [1], epsilon+0.5, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")       
    rowducted_df_26 = krd.rowduct(df, [1], epsilon+1, nb_rows_memory=7000, step_activated = True, \
        label_col = "Label", rowduction_method = "separately")

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)
    assert len(rowducted_df_12) < len(df)
    assert len(rowducted_df_13) < len(df)
    assert len(rowducted_df_14) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_15) <= len(rowducted_df_11)
    assert len(rowducted_df_16) <= len(rowducted_df_15)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_11) == len(rowducted_df_12)
    assert len(rowducted_df_12) == len(rowducted_df_13)
    assert len(rowducted_df_13) == len(rowducted_df_14)
    # In terms of rowducted dataframe
    assert (rowducted_df_11 != rowducted_df_12).any(1).count() == len(rowducted_df_12)
    assert (rowducted_df_12 != rowducted_df_13).any(1).count() == len(rowducted_df_13)
    assert (rowducted_df_13 != rowducted_df_14).any(1).count() == len(rowducted_df_14)

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)
    assert len(rowducted_df_22) < len(df)
    assert len(rowducted_df_23) < len(df)
    assert len(rowducted_df_24) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_25) <= len(rowducted_df_21)
    assert len(rowducted_df_26) <= len(rowducted_df_25)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_21) == len(rowducted_df_22)
    assert len(rowducted_df_22) == len(rowducted_df_23)
    assert len(rowducted_df_23) == len(rowducted_df_24)
    # In terms of rowducted dataframe
    assert (rowducted_df_21 != rowducted_df_22).any(1).count() == len(rowducted_df_22)
    assert (rowducted_df_22 != rowducted_df_23).any(1).count() == len(rowducted_df_23)
    assert (rowducted_df_23 != rowducted_df_24).any(1).count() == len(rowducted_df_24)

    # Check if the dataframes rowductioned "separately" have more or at least the same number
    # of rows than one rowductioned in a "grouped" way
    assert len(rowducted_df_21) >= len(rowducted_df_11)
    assert len(rowducted_df_22) >= len(rowducted_df_12)
    assert len(rowducted_df_23) >= len(rowducted_df_13)
    assert len(rowducted_df_24) >= len(rowducted_df_14)
   
def test_rowduction_01_dataset_a(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_pima_extract.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 0 and 1 values of the label of the Dataset A.

    Implemented checks :
        -Check if the rowduction reduces the number of rows from the original input df
        -Check if all "one shot" rowductions and rowduction with a
        higher batch size (nb_rows_memory) than the total number of rows give the results
        -Check if rowduct(...) and apply_kern_rowduction(...) give same results when
        the parameters are similar
        -Check if it gives the same reduced results with no rowduction_target or equal to "all"
        -Check if it gives the same reduced results with no label_col (nor rowduction_target) \
            and whatever the rowduction_method
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_11bis = krd.rowduct(df, "all", epsilon, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_12 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=73000, \
        step_activated = True, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_13 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_13bis = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_14 = krd.apply_kern_rowduction(df, epsilon, \
        nb_rows_memory=73000).reset_index(drop=True)
    rowducted_df_15 = krd.rowduct(df, "all", epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_16 = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_17 = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=73000, \
        step_activated = False, rowduction_method = "grouped")
    # Since the label_col is by default below, only 1 "separated" value is chosen as target
    rowducted_df_17b = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=73000, \
        step_activated = False, rowduction_method = "separately")
        # With different values of epsilon
    rowducted_df_18 = krd.rowduct(df, [0,1], epsilon+0.05, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "grouped")      
    rowducted_df_19 = krd.rowduct(df, [0,1], epsilon+1, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")
    rowducted_df_21bis = krd.rowduct(df, "all", epsilon, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")
    rowducted_df_22 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=73000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")
    rowducted_df_23 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "separately")
    rowducted_df_23bis = krd.rowduct(df, "all", epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "separately")
    rowducted_df_24a = krd.apply_kern_rowduction(df[df["Label"] == 0].reset_index(drop=True), \
        epsilon, nb_rows_memory=32000)
    rowducted_df_24b = krd.apply_kern_rowduction(df[df["Label"] == 1].reset_index(drop=True), \
        epsilon, nb_rows_memory=32000)
    rowducted_df_24 = rowducted_df_24a.append(rowducted_df_24b).reset_index(drop=True)
    rowducted_df_25 = krd.rowduct(df, rowduction_target="all", epsilon=epsilon, \
        nb_rows_memory=73000, step_activated = False, label_col = "Label", \
        rowduction_method = "separately")
    rowducted_df_26 = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=73000, \
        step_activated = False, label_col = "Label", rowduction_method = "separately")
        # With different values of epsilon
    rowducted_df_27 = krd.rowduct(df, [0,1], epsilon+0.5, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")       
    rowducted_df_28 = krd.rowduct(df, [0,1], epsilon+1, nb_rows_memory=4000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)
    assert len(rowducted_df_11bis) < len(df)
    assert len(rowducted_df_12) < len(df)
    assert len(rowducted_df_13) < len(df)
    assert len(rowducted_df_13bis) < len(df)
    assert len(rowducted_df_14) < len(df)
    assert len(rowducted_df_15) < len(df)
    assert len(rowducted_df_16) < len(df)
    assert len(rowducted_df_17) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_18) <= len(rowducted_df_15)
    assert len(rowducted_df_19) <= len(rowducted_df_18)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_11) == len(rowducted_df_12)
    assert len(rowducted_df_11bis) == len(rowducted_df_12)
    assert len(rowducted_df_11) == len(rowducted_df_11bis)
    assert len(rowducted_df_12) == len(rowducted_df_13)
    assert len(rowducted_df_13) == len(rowducted_df_14)
    assert len(rowducted_df_13bis) == len(rowducted_df_14)
    assert len(rowducted_df_13) == len(rowducted_df_13bis)
    assert len(rowducted_df_14) == len(rowducted_df_15)
    assert len(rowducted_df_15) == len(rowducted_df_16)
    assert len(rowducted_df_16) == len(rowducted_df_17)
    assert len(rowducted_df_17) == len(rowducted_df_17b)

    # In terms of rowducted dataframe
    assert (rowducted_df_11 != rowducted_df_12).any(1).count() == len(rowducted_df_12)
    assert (rowducted_df_11bis != rowducted_df_12).any(1).count() == len(rowducted_df_12)
    assert (rowducted_df_11 != rowducted_df_11bis).any(1).count() == len(rowducted_df_11bis)
    assert (rowducted_df_12 != rowducted_df_13).any(1).count() == len(rowducted_df_13)
    assert (rowducted_df_13 != rowducted_df_14).any(1).count() == len(rowducted_df_14)
    assert (rowducted_df_13bis != rowducted_df_14).any(1).count() == len(rowducted_df_14)
    assert (rowducted_df_13 != rowducted_df_13bis).any(1).count() == len(rowducted_df_13bis)
    assert (rowducted_df_14 != rowducted_df_15).any(1).count() == len(rowducted_df_15)
    assert (rowducted_df_15 != rowducted_df_16).any(1).count() == len(rowducted_df_16)
    assert (rowducted_df_16 != rowducted_df_17).any(1).count() == len(rowducted_df_17)
    assert (rowducted_df_17 != rowducted_df_17b).any(1).count() == len(rowducted_df_17b)


    del rowducted_df_11, rowducted_df_11bis, rowducted_df_12, rowducted_df_13, \
        rowducted_df_13bis, rowducted_df_14, rowducted_df_15, rowducted_df_16, \
        rowducted_df_17, rowducted_df_18, rowducted_df_19

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)
    assert len(rowducted_df_21bis) < len(df)
    assert len(rowducted_df_22) < len(df)
    assert len(rowducted_df_23) < len(df)
    assert len(rowducted_df_23bis) < len(df)
    assert len(rowducted_df_24) < len(df)
    assert len(rowducted_df_25) < len(df)
    assert len(rowducted_df_26) < len(df)
        # If the number of rows decreases when epsilon increases
    assert len(rowducted_df_27) <= len(rowducted_df_26)
    assert len(rowducted_df_28) <= len(rowducted_df_27)

    # Check if the various rowduction parameters lead to a "one shot rowduction"
        # and so the same results
    # In terms of number of rows
    assert len(rowducted_df_21) == len(rowducted_df_22)
    assert len(rowducted_df_21bis) == len(rowducted_df_22)
    assert len(rowducted_df_21) == len(rowducted_df_21bis)
    assert len(rowducted_df_22) == len(rowducted_df_23)
    assert len(rowducted_df_23) == len(rowducted_df_24)
    assert len(rowducted_df_23bis) == len(rowducted_df_24)
    assert len(rowducted_df_23) == len(rowducted_df_23bis)
    assert len(rowducted_df_24) == len(rowducted_df_25)
    assert len(rowducted_df_25) == len(rowducted_df_26)

    # In terms of rowducted dataframe
    assert (rowducted_df_21 != rowducted_df_22).any(1).count() == len(rowducted_df_22)
    assert (rowducted_df_21bis != rowducted_df_22).any(1).count() == len(rowducted_df_22)
    assert (rowducted_df_21 != rowducted_df_21bis).any(1).count() == len(rowducted_df_21bis)
    assert (rowducted_df_22 != rowducted_df_23).any(1).count() == len(rowducted_df_23)
    assert (rowducted_df_23 != rowducted_df_24).any(1).count() == len(rowducted_df_24)
    assert (rowducted_df_23bis != rowducted_df_24).any(1).count() == len(rowducted_df_24)
    assert (rowducted_df_23 != rowducted_df_23bis).any(1).count() == len(rowducted_df_23bis)
    assert (rowducted_df_24 != rowducted_df_25).any(1).count() == len(rowducted_df_25)
    assert (rowducted_df_25 != rowducted_df_26).any(1).count() == len(rowducted_df_26)

    del rowducted_df_21, rowducted_df_21bis, rowducted_df_22, rowducted_df_23, \
        rowducted_df_23bis, rowducted_df_24a, rowducted_df_24b, rowducted_df_24, \
        rowducted_df_25, rowducted_df_26,rowducted_df_27, rowducted_df_28

################## Tests on the Dataset B ##################

def test_rowduction_0_dataset_b(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_adult_extract.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 0 value of the label of the Dataset B. There are less tests than
    with Dataset A because the DF is too heavy to "one shot" rowduct it.

    Implemented checks :
        -Check if the rowduction reduces the number of rows from the original input df
        -Check if the step by step rowduction works and avoid the memory error due to a too large df
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [0], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [0], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "separately")


    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)

    del rowducted_df_11, rowducted_df_21


def test_rowduction_1_dataset_b(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_adult_extract.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 0 value of the label of the Dataset B. There are less tests than
    with Dataset A because the DF is too heavy to "one shot" rowduct it.

    Implemented checks :
        -Check if the rowduction reduces the number of rows from the original input df
        -Check if the step by step rowduction works and avoid the memory error due to a too large df
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [1], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [1], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "separately")

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)

    del rowducted_df_11, rowducted_df_21

def test_rowduction_01_dataset_b(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_adult_extract.csv",
    epsilon = 0.052
    ):
    """Test the rowduction on a 0 and 1 values of the label of the Dataset B. There are less tests
       than with Dataset A because the DF is too heavy to "one shot" rowduct it.

    Implemented checks :
        -Check if the rowduction reduces the number of rows from the original input df
        -Check if the step by step rowduction works and avoid the memory error due to a too large df
    """

    # Load the data
    df = pd.read_csv(path)

    # Rowduct the data in a "grouped" way
    rowducted_df_11 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "grouped")
    rowducted_df_12 = krd.rowduct(df, "all", epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "grouped")

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [0,1], epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "separately")
    rowducted_df_22 = krd.rowduct(df, "all", epsilon, nb_rows_memory=500, step_activated = True,\
        label_col = "Label", rowduction_method = "separately")

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)
    assert len(rowducted_df_12) < len(df)
    assert len(rowducted_df_11) == len(rowducted_df_12)

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)
    assert len(rowducted_df_22) < len(df)
    assert len(rowducted_df_21) == len(rowducted_df_22)

    del rowducted_df_11, rowducted_df_12, rowducted_df_21, rowducted_df_22

################## Tests on the Dataset C ##################

def test_rowduction_dataset_c(
    path = os.path.dirname(os.path.realpath(__file__)) + "/data/clean_abalone.csv",
    epsilon = 0.05
    ):
    """Test the rowduction on a 0 and 1 values of the label of the Dataset B. There are less tests
       than with Dataset A because the DF is too heavy to "one shot" rowduct it.

    Implemented checks :
        -Check if the rowduction reduces the number of rows from the original input df
        -Check if the rowduction for a multiclass dataset reduces the rows with the given \
            rowduction_target but not the others
       -Check if the rowduction with all the classes as target (and even more than the ones that \
           actually exist) is equivalent to an "all" rowduction
    """

    # Load the data
    df = pd.read_csv(path)
    print(df["Label"])

    # Rowduct the data in a "grouped" way
        # On specific targets
    rowducted_df_11 = krd.rowduct(df, [10, 11, 12], epsilon, nb_rows_memory=10000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_12a = krd.apply_kern_rowduction(\
        df[df["Label"].isin([10,11,12])].reset_index(drop=True), epsilon, nb_rows_memory=10000
        )
    rowducted_df_12b = df[~df["Label"].isin([10,11,12])]
    rowducted_df_12 = rowducted_df_12a.append(rowducted_df_12b).sample(frac=1)
        # On all targets
    rowducted_df_13 = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=10000, \
        step_activated = False, label_col = "Label", rowduction_method = "grouped")
    rowducted_df_14 = krd.rowduct(df, rowduction_target=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, \
        12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29], \
        epsilon=epsilon, nb_rows_memory=10000, step_activated = False, label_col = "Label", \
        rowduction_method = "grouped")

    # Test the "grouped" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_11) < len(df)
    assert len(rowducted_df_12) < len(df)
    assert len(rowducted_df_13) < len(df)
    assert len(rowducted_df_14) < len(df)
    # Check if the "targeting" works as independent rowductions
    assert len(rowducted_df_11) == len(rowducted_df_12)
    assert len(rowducted_df_11[rowducted_df_11["Label"].isin([10,11,12])]) == len(rowducted_df_12a)
    assert len(rowducted_df_11[~rowducted_df_11["Label"].isin([10,11,12])]) == len(rowducted_df_12b)
     # Check if the rowductions remains the same whether all classes
     # or "all" are the rowduction_target
    assert len(rowducted_df_13) == len(rowducted_df_14)

    del rowducted_df_11, rowducted_df_12a, rowducted_df_12b, rowducted_df_12, rowducted_df_13, \
        rowducted_df_14

    # Rowduct the data "separately"
    rowducted_df_21 = krd.rowduct(df, [10, 11, 12], epsilon, nb_rows_memory=10000, \
        step_activated = True, label_col = "Label", rowduction_method = "separately")
    rowducted_df_22a = krd.apply_kern_rowduction(df[df["Label"] == 10].reset_index(drop=True), \
        epsilon, nb_rows_memory=10000)
    rowducted_df_22b = krd.apply_kern_rowduction(df[df["Label"] == 11].reset_index(drop=True), \
        epsilon, nb_rows_memory=10000)
    rowducted_df_22c = krd.apply_kern_rowduction(df[df["Label"] == 12].reset_index(drop=True), \
        epsilon, nb_rows_memory=10000)
    rowducted_df_22d = df[~df["Label"].isin([10,11,12])]
    rowducted_df_22 = rowducted_df_22a.append(rowducted_df_22b).append(
        rowducted_df_22c).append(rowducted_df_22d).sample(frac=1)
    rowducted_df_23 = krd.rowduct(df, epsilon=epsilon, nb_rows_memory=10000, \
        step_activated = False, label_col = "Label", rowduction_method = "separately")
    rowducted_df_24 = krd.rowduct(df, rowduction_target=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, \
        11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29], \
        epsilon=epsilon, nb_rows_memory=10000, step_activated = False, label_col = "Label", \
        rowduction_method = "separately")

    # Test the "separately" Rowduction

    # Check if the rowduction reduces the number of rows
    assert len(rowducted_df_21) < len(df)
    assert len(rowducted_df_22) < len(df)
    assert len(rowducted_df_23) < len(df)
    assert len(rowducted_df_24) < len(df)
    # Check if the "targeting" works as independent rowductions
    assert len(rowducted_df_21) == len(rowducted_df_22)
    assert len(rowducted_df_21[rowducted_df_21["Label"].isin([10,11,12])]) == len(rowducted_df_22a)\
        + len(rowducted_df_22b) + len(rowducted_df_22c)
    assert len(rowducted_df_21[~rowducted_df_21["Label"].isin([10,11,12])]) == len(rowducted_df_22d)
     # Check if the rowductions remains the same whether all classes
     # or "all" are the rowduction_target
    assert len(rowducted_df_23) == len(rowducted_df_24)

    del rowducted_df_21, rowducted_df_22a, rowducted_df_22b, rowducted_df_22c, rowducted_df_22d, \
        rowducted_df_22, rowducted_df_23, rowducted_df_24

################## Tests on the Dataset D ##################

def test_rowduction_dataset_d():
    """Test the rowduction on a 0 and 1 values of the label of a Dataset D.

    Implemented checks :
        -Check if the value of epsilon impacts the delete of isolated points
    """

    data = {'col1': [14, 10, 10, 10, 10, 10], 'col2': [15, 15, 50, 15, 15, 15], 'col3': [20, 20, 20, 20, 20, 20],
     'col4': [25, 30, 25, 25, 25, 25], 'label': [1, 1, 1, 0, 0, 0]}
    df = pd.DataFrame(data = data)
    
    # Epsilon low to keep isolated points
    rowductioned_df_1 = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.1,label_col='label',rowduction_method='separately',remove_isolated_points=False)
    rowductioned_df_1b = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.1,label_col='label',rowduction_method='separately',remove_isolated_points=True)

    rowductioned_df_2 = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.4,label_col='label',rowduction_method='separately',remove_isolated_points=False)
    rowductioned_df_2b = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.4,label_col='label',rowduction_method='separately',remove_isolated_points=True)

    # Epsilon higher to remove isolated points
    rowductioned_df_3 = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.5,label_col='label',rowduction_method='separately',remove_isolated_points=False)
    rowductioned_df_3b = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.5,label_col='label',rowduction_method='separately',remove_isolated_points=True)

    rowductioned_df_4 = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.7,label_col='label',rowduction_method='separately',remove_isolated_points=False)
    rowductioned_df_4b = krd.rowduct(df,rowduction_target=[0,1],\
        epsilon=0.7,label_col='label',rowduction_method='separately',remove_isolated_points=True)

    # Check that 
        # if epsilon < 0.45 we consider df's rows of index 0 & 1 as isolated points --> keep
        # if epsilon >= 0.5 we DON'T consider df's rows of index 0 & 1 as isolated points --> remove
    assert rowductioned_df_1.equals(rowductioned_df_2) == True
    assert rowductioned_df_2.equals(rowductioned_df_3) == False
    assert rowductioned_df_3.equals(rowductioned_df_4) == True

    # Check if the first row is found in the reduced dataframes with epsilon < 0.45 but not if 
    # epsilon >= 0.5
    assert ((rowductioned_df_1['col1'] == 14) & (rowductioned_df_1['col2'] == 15) \
                & (rowductioned_df_1['col3'] == 20) & (rowductioned_df_1['col4'] == 25) \
                & (rowductioned_df_1['label'] == 1)
                ).any() == True
    assert ((rowductioned_df_2['col1'] == 14) & (rowductioned_df_2['col2'] == 15) \
                & (rowductioned_df_2['col3'] == 20) & (rowductioned_df_2['col4'] == 25) \
                & (rowductioned_df_2['label'] == 1)
                ).any() == True

    assert ((rowductioned_df_3['col1'] == 14) & (rowductioned_df_3['col2'] == 15) \
                & (rowductioned_df_3['col3'] == 20) & (rowductioned_df_3['col4'] == 25) \
                & (rowductioned_df_3['label'] == 1)
                ).any() == False
    assert ((rowductioned_df_3['col1'] == 14) & (rowductioned_df_3['col2'] == 15) \
                & (rowductioned_df_3['col3'] == 20) & (rowductioned_df_3['col4'] == 25) \
                & (rowductioned_df_3['label'] == 1)
                ).any() == False

    # Check if the isolated point are removed when remove_isolated_points is True
    assert ((rowductioned_df_1b['col1'] == 14) & (rowductioned_df_1b['col2'] == 15) \
                & (rowductioned_df_1b['col3'] == 20) & (rowductioned_df_1b['col4'] == 25) \
                & (rowductioned_df_1b['label'] == 1)
                ).any() == False
    assert ((rowductioned_df_2b['col1'] == 14) & (rowductioned_df_2b['col2'] == 15) \
                & (rowductioned_df_2b['col3'] == 20) & (rowductioned_df_2b['col4'] == 25) \
                & (rowductioned_df_2b['label'] == 1)
                ).any() == False

    assert ((rowductioned_df_3b['col1'] == 14) & (rowductioned_df_3b['col2'] == 15) \
                & (rowductioned_df_3b['col3'] == 20) & (rowductioned_df_3b['col4'] == 25) \
                & (rowductioned_df_3b['label'] == 1)
                ).any() == False
    assert ((rowductioned_df_4b['col1'] == 14) & (rowductioned_df_4b['col2'] == 15) \
                & (rowductioned_df_4b['col3'] == 20) & (rowductioned_df_4b['col4'] == 25) \
                & (rowductioned_df_4b['label'] == 1)
                ).any() == False 


"""The setup.py of the Kern_Rowduction package."""

from setuptools import setup, find_packages
from pathlib import Path


VERSION = '0.0.4'
DESCRIPTION = 'Kern Rowduction - A package to reduce the number of rows / undersample the \
    (imbalanced) datasets by graph kernelisation methods.'
CURRENT_DIRECTORY = Path(__file__).parent
LONG_DESCRIPTION = (CURRENT_DIRECTORY / "README.md").read_text()

# Setting up
setup(
        name="Kern-Rowduction",
        version=VERSION,
        author="Hichem Boughattas & Hamza Bouanani",
        author_email="kern.rowduction@gmail.com",
        description=DESCRIPTION,
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        packages=find_packages(),
        install_requires=["pandas","numpy","networkx"],
        url = 'https://github.com/Kern-Rowduction/Kern-Rowduction',
        download_url = 'https://github.com/Kern-Rowduction/Kern-Rowduction/archive/refs/tags/v_0.0.4.tar.gz',
        keywords=['python', 'graph', "kernel", "rowduction", \
        "imbalance", "data", "science", "dataset"],
        classifiers= [
            "Development Status :: 3 - Alpha",
            "Intended Audience :: Developers",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 3",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "Topic :: Scientific/Engineering :: Mathematics"
        ]
)



"""All the usable methods/functions which constitute the Kern_Rowduction package."""

import pandas as pd
import numpy as np
import networkx as nx
from tqdm import tqdm

def epsilon_dominates(point_id_1=int, point_id_2=int, df=pd.DataFrame, epsilon=float):
    '''From an input DF, check if the input node Point 1 (row 1) epsilon dominates the node
        Point 2 (row 2).

    Parameters
    ----------
    point_id_1 : int
        Row index of the point 1 in the pandas DataFrame df
    point_id_2 : int
        Row index of the point 2 in the pandas DataFrame df
    df : pd.DataFrame
        Pandas Dataframe where the point are stored
    epsilon : float
        Threshold of dominance between 2 points: Point1 must be > (1+epsilon)*Point2 for each column

    Returns
    -------
    flag
        A boolean which is True if Point 1 epsilon dominates Point 2 otherwise False
    '''

    flag = True #Assumed true unless evidence to the contrary

    # Get all the edged nodes of Point 1 & 2
    point_1 = list(df.iloc[point_id_1])
    point_2 = list(df.iloc[point_id_2])

    # Iteration over each edged node of Point 1
    for i in range(len(point_1)):
        if point_1[i] > (1+epsilon)*point_2[i]:
            flag = False
            break

    return flag # True if Point 1 epsilon dominates Point 2 otherwise False

def find_in_neighbors(graph, point_id, all_neighbors_dic):
    '''From a graph and a dictionary of all nodes' neighbors, find all INNER neighbors of
       the input node (identified by its point_id).

    Parameters
    ----------
    graph : networkx DiGraph
        Networkx Digraph with all the nodes and edges created basically from the original DataFrame
    point_id : int
        Number ID of the node to analyze in the graph
    all_neighbors_dic : dict
        Dictionary of integers (nodes) with all the inner and outer neighbors of the all the nodes

    Returns
    -------
    in_neighbors : list
        A list with all the INNER neighbords of the node point_id
    '''
    all_neighbors = all_neighbors_dic[point_id] # Neighbors of point_id
    in_neighbors = []
    # For all the edged node IDs in the dictionary for the input node
    for neighbor in all_neighbors:
        #If there is an edge TO the input node, append it to the list of all INNER neighbors
        if (neighbor, point_id) in graph.edges():
            in_neighbors.append(neighbor)
    return in_neighbors

def find_out_neighbors(graph, point_id, all_neighbors_dic):
    '''From a graph and a dictionary of all nodes' neighbors, find all OUTER neighbors of
       the input node (identified by its point_id).

    Parameters
    ----------
    graph : networkx DiGraph
        Networkx Digraph with all the nodes and edges created basically from the original DataFrame
    point_id : int
        Number ID of the node to analyze in the graph
    all_neighbors_dic : dict
        Dictionary of integers (nodes) with all the inner and outer neighbors of the all the nodes

    Returns
    -------
    in_neighbors : list
        A list with all the OUTER neighbords of the node point_id
    '''
    all_neighbors = all_neighbors_dic[point_id] # Neighbors of point_id
    out_neighbors = []
    # For all the edged node IDs in the dictionary for the input node
    for neighbor in all_neighbors:
        #If there is an edge FROM the input node, append it to the list of all OUTER neighbors
        if (point_id, neighbor) in graph.edges():
            out_neighbors.append(neighbor)
    return out_neighbors

def extract_kernel(graph, all_neighbors_dic):
    '''Retrieve the kernel of the input directed graph if there is one.
        NB : This function works indefinitely if the graph includes a circuit.

    Parameters
    ----------
    graph : networkx DiGraph
        Networkx Digraph with all the nodes and edges created basically from the original DataFrame
    all_neighbors_dic : dict
        Dictionary of integers (nodes) with all the inner and outer neighbors of the all the nodes

    Returns
    -------
    kernel : list
        A sorted list with all the nodes of the kernel computed.
    '''
    # Initialize variables
    kernel = []
    nodes = graph.nodes()

    # Initialize list of nodes to check
    d_minus, candidate = {}, {}
    for i in nodes:
        d_minus[i] = 0
        candidate[i] = False

    # Iteration over each node of the graph
    for x in nodes:
        d_minus[x] = len(find_in_neighbors(graph, x, all_neighbors_dic))
        if d_minus[x] == 0: # If the node has no previous one, it's added to the kernel
            kernel.append(x)
        else: # Otherwise it is added as a potential candidate to check as kernel's
                # node by the search algorithm
            candidate[x] = True

    nb_candidates = sum([candidate[key] for key in candidate])

    # Search algorithm : Check if the candidates are kernel's node or not
    # So long there is still at least 1 remaining candidate node
    while nb_candidates > 0:

        # For each node x in the kernel
        for x in kernel:
            # For each OUT neighbors of this node X
            for y in find_out_neighbors(graph, x, all_neighbors_dic):

                # If the found OUT neighbor node Y is also a candidate ==> we remove it
                    # from the list of candidates
                if candidate[y]:
                    candidate[y] = False
                    nb_candidates -= 1

                    # For each OUT neighbors of this IN neighbor node Y
                    for z in find_out_neighbors(graph, y, all_neighbors_dic):

                        if candidate[z]: # If the found OUT neighbor node Z of Y
                                            # is a remaining candidate, remove it
                            d_minus[z] -= 1
                            if d_minus[z] == 0: # If it's the last node, then there is no
                                                    # following nodes, so append it to the kernel
                                kernel.append(z)
                                candidate[z] = False
                                nb_candidates -= 1

    # Return the final computed kernel
    return sorted(kernel)

def build_uncycled_subgraphs(graph):
    '''Build 2 induced subgraphs from the input graph so that both subgraphs don't include cycles.
       That is to say, from a input graph, build 2 sub graphs according to the sense of their edges
       (departure_node's value <=> arrival_node's value).

    Parameters
    ----------
    graph : networkx DiGraph
        Networkx Digraph with all the nodes and edges created basically from the original DataFrame

    Returns
    -------
    graph1_prime : networkx DiGraph
        A uncycled sub graph with the edges between a lower node's value to a higher node's value
        from the input graph
    graph2_prime : networkx DiGraph
        A uncycled sub graph with the other edges from the input graph not put into graph1_prime
    '''

    # Make the 2 subgraphs with the same nodes
    graph1_prime = nx.DiGraph()
    graph1_prime.add_nodes_from(list(graph.nodes))
    graph2_prime = nx.DiGraph()
    graph2_prime.add_nodes_from(list(graph.nodes))

    # Copy the edges of the input graph in one or another subgraph according to the values
        # of the entry/exit nodes
    for e in graph.edges():
        if e[0] < e[1]: # If the entry node's value < exit node's value ==> add the
                            # edge to the subgraph 1
            graph1_prime.add_edge(e[0], e[1])
        else: # Otherwise ==> add the edge of the subgraph 2
            graph2_prime.add_edge(e[0], e[1])

    return graph1_prime, graph2_prime

def extract_quasi_kernel(uncycled_subgraph_1, uncycled_subgraph_2, all_neighbors_dic):
    '''Extract the quasi kernel from 2 associated uncycled subgraphs.

    Parameters
    ----------
    uncycled_subgraph_1 : networkx DiGraph
        Networkx Digraph without cycles and with all the nodes and "increasing" edges extracted
        initially from a bigger input graph.
    uncycled_subgraph_2 : networkx DiGraph
        Networkx Digraph without cycles and with all the nodes and "decreasing" edges extracted
        initially from a bigger input graph.
    all_neighbors_dic : dict
        Dictionary of integers (nodes) with all the inner and outer neighbors of the all the nodes

    Returns
    -------
    quasi_kernel : list
        A sorted list with all the nodes of the quasi kernel computed.
    '''

    subgraph_2_kernel = extract_kernel(uncycled_subgraph_1, all_neighbors_dic)
    subgraph_2_prime = uncycled_subgraph_2.subgraph(subgraph_2_kernel)
    quasi_kernel = extract_kernel(subgraph_2_prime, all_neighbors_dic)
    return quasi_kernel

def apply_kern_rowduction(df, epsilon, nb_rows_memory=100000, remove_isolated_points=False):
    '''Apply the "Rowduction" process on a DF and returns the reduced DataFrame.
        NB : The Rowduction process consists in building a directed graph, extracting its quasi
        kernel and returning the associated DF.

    Parameters
    ----------
    df : pandas DataFrame
        DataFrame to reduce through the "Rowduction" process
    epsilon : float
        Threshold of dominance between 2 points: Point1 must be > (1+epsilon)*Point2 for each column
    nb_rows_memory : int
        Number of rows per batch to process the Rowduction. Smaller it is, less memory it takes but
        potentially slower it is
    remove_isolated_points : boolean
        Enabled/Disabled the remove of the 'isolated' points/nodes in the graph, that is to say
        the nodes for which no epsilon dominance relationship has been found.

    Returns
    -------
    rowducted_df : pandas DataFrame
        The input 'df' DataFrame but rowducted, with only the rows to keep after the Rowduction
    '''

    # Initialize an empty DiGraph and all variables
    graph = nx.DiGraph() # Creation of a directed graph
    edges_list = []
    nodes_list = [row_index for row_index in df.index] # Nodes to add in the graph

    all_neighbors_dic = {} #Dictionary of nodes' neighbors

    # Compare each row of df with all of eps_df
    epsilon_prime = -1 + (1+epsilon)**0.5
    eps_df = (1+epsilon_prime)*df
    eps_df_array = np.array(eps_df.values[:, None])

    # For each subframe of the dataframe, we retrieve all epsilon dominated/dominating nodes
        # in a list of dict
    nb_iter = (len(df) // nb_rows_memory) + 1
    df_res = pd.DataFrame([], columns=['A', 'B'])
    for i in range(0, nb_iter):
        df_array = np.array(df.loc[i*nb_rows_memory : (i+1)*nb_rows_memory, :].values)
        all_greater = (eps_df_array >= df_array).all(axis=2) # It works because of the broadcast
                                                                # property of numpy arrays
        # Get the row indexes of df and eps_df where there is a epsilon dominance
            # thanks nonzero and transpose
        trans_greater = np.transpose(np.nonzero(all_greater))
        trans_greater[:, 1] += i*nb_rows_memory
        df_i = pd.DataFrame(trans_greater, columns=['A', 'B'])
        df_res = df_res.append(df_i)
        # Remove the self indexes (0 epsilon dominates 0 etc...)
        # And put into a dictionary the row indexes which are epsilon dominating or dominated
        df_res = df_res.drop_duplicates().reset_index(drop=True)
        dict_dominating = df_res[df_res['A'].ne(df_res['B'])].groupby('A')['B'].agg(list).to_dict()
        dict_dominated = df_res[df_res['A'].ne(df_res['B'])].groupby('B')['A'].agg(list).to_dict()

    # Merge both dominating & dominated nodes dictionaries + create the list of edges (as tuples)
    for k in df.index:
        node_doms = []
        edges = []

        if k in dict_dominated:
            node_doms.extend(dict_dominated[k])

            # Add edges to the edges_list
            edges = list(map(lambda x: (k, x), dict_dominated[k]))
            edges_list.extend(edges)

        if k in dict_dominating:
            node_doms.extend(dict_dominating[k])

        # Remove duplicates and add to the neighbor dictionary all merged dominated/dominating nodes
        node_doms = list(dict.fromkeys(node_doms))
        node_doms.sort()
        all_neighbors_dic[k] = node_doms

        del node_doms, edges

    # Build the global graph

        # Add all nodes
    if not remove_isolated_points:
        graph.add_nodes_from(nodes_list) # Can be deleted if you wish to remove 'pseudo outliers'

        # Add all edges of dominance relationships
    graph.add_edges_from(edges_list)

    # Compute quasi kernel
    graph_1_prime, graph_2_prime = build_uncycled_subgraphs(graph)
    quasi_kernel = extract_quasi_kernel(graph_1_prime, graph_2_prime, all_neighbors_dic)

    rowducted_df = df.iloc[quasi_kernel]
    return rowducted_df

def rowduct(df=pd.DataFrame, rowduction_target="all", epsilon=0.025, nb_rows_memory=10000, \
    step_activated=False, label_col="_krd_default_label_", rowduction_method="separately", \
    remove_isolated_points=False, shuffle_seed=42):
    '''Parametrize the "Rowduction" process, apply it on a DF and returns the reduced DataFrame.

    Parameters
    ----------
    df : pandas DataFrame
        DataFrame to reduce through the "Rowduction" process
    rowduction_target : list or str
        List of label values upon which to apply the Rowduction process. Rows with label values
        that are not listed will remain intact. If the rowduction_target = "all" (default value),
        the whole input DataFrame df is reduced.
    epsilon : float
        Threshold of dominance between 2 points: Point1 must be > (1+epsilon)*Point2 for each column
    nb_rows_memory : int
        Number of rows per batch to process the Rowduction. Smaller it is, less memory it takes but
        potentially slower it is
    step_activated : boolean
        True if the Rowduction has to be step by step of rows, which avoids memory errors in case of
        huge input DataFrame df. Otherwise False for a 'one shot' rowduction.
    label_col : str
        Name of the label column which contains the 'rowduction_target'
    rowduction_method : str
        Rowduction method : 'grouped' to reduce/rowduct together all Rowduction targets,
        'separately' to reduce/rowduct each of them separately and then concatenate them together
    remove_isolated_points : boolean
        Enabled/Disabled the remove of the 'isolated' points/nodes in the graph, that is to say
        the nodes for which no epsilon dominance relationship has been found.7
    shuffle_seed : int
        A seed for the random shuffles of the rows done during the Kern Rowduction process. If 
        the seed is set (42 by default), the order of the rows returned at the end of the 
        Kern Rowduction will remain unchanged.

    Returns
    -------
    rowducted_df : pandas DataFrame
        The input 'df' DataFrame but rowducted, with only the rows to keep after the Rowduction
    '''

    # If label_col is equal to the default value, \
    # then create a tempory constant column "_krd_default_label_"
    if label_col == "_krd_default_label_":
        df["_krd_default_label_"] = 0

    # If the rowduction target is "all", the target are the list of distinct values of the label_col
    if rowduction_target == "all":
        rowduction_target = list(set(df[label_col]))

    header = df.columns
    nb_iter = (len(df) // nb_rows_memory) + 1

    # If we want to do a step by step rowduction
    if step_activated and nb_iter > 1:


        # Initialize the dataframe concatenating the results of the various rowductions
        all_rd_data = pd.DataFrame([], columns=header)

        # For each batch of data with initialization the progress bar displayed in the console
        for i in tqdm(range(nb_iter), desc='... Kern Rowduction - Progression (by batch)'):

            index_begin = i*nb_rows_memory
            index_end = (i+1)*nb_rows_memory-1

            # Rowduct all values of the label together (0 and 1 together by example)
            if rowduction_method == "grouped":
                data = df.loc[index_begin:index_end].reset_index(drop=True)
                data_target = data[data[label_col].isin(rowduction_target)].reset_index(drop=True)
                rd_data_target = apply_kern_rowduction(data_target, epsilon,\
                    nb_rows_memory=nb_rows_memory, \
                    remove_isolated_points=remove_isolated_points)
                data_rest = data[~data[label_col].isin(rowduction_target)].reset_index(drop=True)
                rd_data = rd_data_target.append(data_rest).sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)
                all_rd_data = all_rd_data.append(rd_data)
                del rd_data, data, data_target, data_rest, rd_data_target


            # Rowduct all values of the label separately, and then join them together
                # with the none rowducted ones
            elif rowduction_method == "separately":
                data = df.loc[index_begin:index_end].reset_index(drop=True)
                data_target = data[data[label_col].isin(rowduction_target)].reset_index(drop=True)
                data_rest = data[~data[label_col].isin(rowduction_target)].reset_index(drop=True)
                rd_data = data_rest.copy()
                for target in rowduction_target:
                    iter_data_target = data_target[data_target[label_col] == target]
                    iter_data_target = iter_data_target.reset_index(drop=True)
                    iter_rd_data_target = apply_kern_rowduction(iter_data_target, epsilon, \
                        nb_rows_memory=nb_rows_memory, \
                        remove_isolated_points=remove_isolated_points)
                    rd_data = rd_data.append(iter_rd_data_target)
                rd_data = rd_data.sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)
                all_rd_data = all_rd_data.append(rd_data)
                del rd_data, data, data_target, data_rest, iter_data_target, iter_rd_data_target

            else:
                return "Error : bad rowduct_target input."


        # At the end rowduct all the rowducted data once again

        all_rd_data = all_rd_data.reset_index(drop=True)

        # Rowduct all values of the label together (0 and 1 together by example)
        if rowduction_method == "grouped":
            all_data_target = all_rd_data[all_rd_data[label_col].isin(rowduction_target)]
            all_data_target = all_data_target.reset_index(drop=True)
            all_rd_data_target = apply_kern_rowduction(all_data_target, epsilon, \
                nb_rows_memory=nb_rows_memory,\
                remove_isolated_points=remove_isolated_points)
            all_data_rest = all_rd_data[~all_rd_data[label_col].isin(rowduction_target)]
            all_data_rest = all_data_rest.reset_index(drop=True)
            final_rd_data = all_rd_data_target.append(all_data_rest)
            final_rd_data = final_rd_data.sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)

            # If label_col is equal to the default value, \
            # then delete the temporary column "_krd_default_label_" previously created
            if label_col == "_krd_default_label_":
                final_rd_data = final_rd_data.drop('_krd_default_label_', axis=1)

            return final_rd_data

        # Rowduct all values of the label separately, and then join them together
            # with the none rowducted ones
        if rowduction_method == "separately":
            all_data_target = all_rd_data[all_rd_data[label_col].isin(rowduction_target)]
            all_data_target = all_data_target.reset_index(drop=True)
            all_data_rest = all_rd_data[~all_rd_data[label_col].isin(rowduction_target)]
            all_data_rest = all_data_rest.reset_index(drop=True)
            final_rd_data = all_data_rest.copy()

            for target in rowduction_target:
                iter_data_target = all_data_target[all_data_target[label_col] == target]
                iter_data_target = iter_data_target.reset_index(drop=True)
                iter_rd_data_target = apply_kern_rowduction(iter_data_target, epsilon,\
                    nb_rows_memory=nb_rows_memory,\
                    remove_isolated_points=remove_isolated_points)
                final_rd_data = final_rd_data.append(iter_rd_data_target)
                final_rd_data = final_rd_data.sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)

            # If label_col is equal to the default value, \
            # then delete the temporary column "_krd_default_label_" previously created
            if label_col == "_krd_default_label_":
                final_rd_data = final_rd_data.drop('_krd_default_label_', axis=1)

            return final_rd_data

    # Else rowduct the whole dataset in one batch
    else:

        print('... Kern Rowduction - In Progression (1 batch) ...')

        # Rowduct all values of the label together (0 and 1 together by example)
        if rowduction_method == "grouped":
            data_target = df[df[label_col].isin(rowduction_target)].reset_index(drop=True)
            rd_data_target = apply_kern_rowduction(data_target, epsilon,\
                nb_rows_memory=nb_rows_memory,\
                remove_isolated_points=remove_isolated_points)
            data_rest = df[~df[label_col].isin(rowduction_target)].reset_index(drop=True)
            rd_data = rd_data_target.append(data_rest)
            rd_data = rd_data.sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)

            # If label_col is equal to the default value, \
            # then delete the temporary column "_krd_default_label_" previously created
            if label_col == "_krd_default_label_":
                rd_data = rd_data.drop('_krd_default_label_', axis=1)

            return rd_data

        # Rowduct all values of the label separately, and then join them together
            # with the none rowducted ones
        if rowduction_method == "separately":
            data_target = df[df[label_col].isin(rowduction_target)].reset_index(drop=True)
            data_rest = df[~df[label_col].isin(rowduction_target)].reset_index(drop=True)
            rd_data = data_rest.copy()

            for target in rowduction_target:
                iter_data_target = data_target[data_target[label_col] == target]
                iter_data_target = iter_data_target.reset_index(drop=True)
                iter_rd_data_target = apply_kern_rowduction(iter_data_target, epsilon,\
                    nb_rows_memory=nb_rows_memory, \
                    remove_isolated_points=remove_isolated_points)
                rd_data = rd_data.append(iter_rd_data_target)
            rd_data = rd_data.sample(frac=1, random_state=shuffle_seed).reset_index(drop=True)

            # If label_col is equal to the default value, \
            # then delete the temporary column "_krd_default_label_" previously created
            if label_col == "_krd_default_label_":
                rd_data = rd_data.drop('_krd_default_label_', axis=1)

            return rd_data

        else:
            return "Error : bad rowduct_target input."
